/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

/**
 * SECTION:shell-mag-inverse-effect
 * @short_description: Inverse lightness effect for the magnifier.
 * @see_also: #ClutterEffect, #ClutterOffscreenEffect
 *
 * #ShellMagInverseEffect is a sub-class of #ClutterEffect that enhances the
 * the appearance of a clutter actor.  Specifically it inverts the lightness
 * of an actor (e.g., darker colors become lighter, white becomes black, and
 * vice versa).
 */

#include "shell-mag-inverse-effect.h"

#include <string.h>
#include <cogl/cogl.h>

struct _ShellMagInverseEffect
{
  ClutterOffscreenEffect parent_instance;

  /* a back pointer to our actor, so that we can query it */
  ClutterActor *actor;

  CoglHandle shader;
  CoglHandle program;

  gint tex_uniform;

  guint is_attached : 1;
};

struct _ShellMagInverseEffectClass
{
  ClutterOffscreenEffectClass parent_class;
};

/* Lightness inversion in GLSL.
 */
static const gchar *magnifier_invert_glsl_shader =
"uniform sampler2D tex;\n"
"\n"
"void main ()\n"
"{\n"
"  vec4 color = cogl_color_in * texture2D (tex, vec2 (cogl_tex_coord_in[0].xy));\n"
"  vec3 effect = vec3 (color);\n"
"\n"
"  float maxColor = max (color.r, max (color.g, color.b));\n"
"  float minColor = min (color.r, min (color.g, color.b));\n"
"  float lightness = (maxColor + minColor) / 2.0;\n"
"\n"
"  float delta = (1.0 - lightness) - lightness;\n"
"  effect.rgb = (effect.rgb + delta);\n"
"\n"
"  cogl_color_out = vec4 (effect, color.a);\n"
"}\n";

G_DEFINE_TYPE (ShellMagInverseEffect,
               shell_mag_inverse_effect,
               CLUTTER_TYPE_OFFSCREEN_EFFECT);

static gboolean
shell_mag_inverse_effect_pre_paint (ClutterEffect *effect)
{
  ShellMagInverseEffect *self = SHELL_MAG_INVERSE_EFFECT (effect);
  ClutterEffectClass *parent_class;

  if (self->shader == COGL_INVALID_HANDLE)
    return FALSE;

  if (!clutter_actor_meta_get_enabled (CLUTTER_ACTOR_META (effect)))
    return FALSE;

  if (self->actor == NULL)
    return FALSE;

  if (self->program == COGL_INVALID_HANDLE)
    self->program = cogl_create_program ();

  if (!self->is_attached)
    {
      g_assert (self->shader != COGL_INVALID_HANDLE);
      g_assert (self->program != COGL_INVALID_HANDLE);

      cogl_program_attach_shader (self->program, self->shader);
      cogl_program_link (self->program);

      cogl_handle_unref (self->shader);

      self->is_attached = TRUE;

      self->tex_uniform =
        cogl_program_get_uniform_location (self->program, "tex");
    }

  parent_class = CLUTTER_EFFECT_CLASS (shell_mag_inverse_effect_parent_class);
  return parent_class->pre_paint (effect);
}

static void
shell_mag_inverse_effect_paint_target (ClutterOffscreenEffect *effect)
{
  ShellMagInverseEffect *self = SHELL_MAG_INVERSE_EFFECT (effect);
  ClutterOffscreenEffectClass *parent;
  CoglHandle material;

  if (self->program == COGL_INVALID_HANDLE)
    goto out;

  if (self->tex_uniform > -1)
    cogl_program_set_uniform_1i (self->program, self->tex_uniform, 0);

  material = clutter_offscreen_effect_get_target (effect);
  cogl_material_set_user_program (material, self->program);

out:
  parent = CLUTTER_OFFSCREEN_EFFECT_CLASS (shell_mag_inverse_effect_parent_class);
  parent->paint_target (effect);
}

static void
shell_mag_inverse_effect_set_actor (ClutterActorMeta *meta,
                                    ClutterActor *actor)
{
  ShellMagInverseEffect *self = SHELL_MAG_INVERSE_EFFECT (meta);
  ClutterActorMetaClass *parent;

  g_return_if_fail (actor == NULL || CLUTTER_IS_ACTOR (actor));

  if (self->shader == COGL_INVALID_HANDLE)
    {
      clutter_actor_meta_set_enabled (meta, FALSE);
      return;
    }

  parent = CLUTTER_ACTOR_META_CLASS (shell_mag_inverse_effect_parent_class);
  parent->set_actor (meta, actor);

  /* keep a back pointer here, to avoid going through the ActorMeta */
  self->actor = clutter_actor_meta_get_actor (meta);
}

static void
shell_mag_inverse_effect_dispose (GObject *gobject)
{
  ShellMagInverseEffect *self = SHELL_MAG_INVERSE_EFFECT (gobject);

  if (self->program != COGL_INVALID_HANDLE)
    {
      cogl_handle_unref (self->program);

      self->program = COGL_INVALID_HANDLE;
      self->shader = COGL_INVALID_HANDLE;
    }

  self->actor = NULL;

  G_OBJECT_CLASS (shell_mag_inverse_effect_parent_class)->dispose (gobject);
}

static void
shell_mag_inverse_effect_set_property (GObject      *gobject,
                                       guint        prop_id,
                                       const GValue *value,
                                       GParamSpec   *pspec)
{
  switch (prop_id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
shell_mag_inverse_effect_get_property (GObject    *gobject,
                                       guint       prop_id,
                                       GValue     *value,
                                       GParamSpec *pspec)
{
  switch (prop_id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
shell_mag_inverse_effect_class_init (ShellMagInverseEffectClass *klass)
{
  ClutterEffectClass *effect_class = CLUTTER_EFFECT_CLASS (klass);
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  ClutterOffscreenEffectClass *offscreen_class;
  ClutterActorMetaClass *meta_class = CLUTTER_ACTOR_META_CLASS (klass);

  offscreen_class = CLUTTER_OFFSCREEN_EFFECT_CLASS (klass);
  offscreen_class->paint_target = shell_mag_inverse_effect_paint_target;

  meta_class->set_actor = shell_mag_inverse_effect_set_actor;

  effect_class->pre_paint = shell_mag_inverse_effect_pre_paint;

  gobject_class->set_property = shell_mag_inverse_effect_set_property;
  gobject_class->get_property = shell_mag_inverse_effect_get_property;
  gobject_class->dispose = shell_mag_inverse_effect_dispose;
}

static void
shell_mag_inverse_effect_init (ShellMagInverseEffect *self)
{
  static CoglHandle shader = COGL_INVALID_HANDLE;

  if (shader == COGL_INVALID_HANDLE)
    {
      if (clutter_feature_available (CLUTTER_FEATURE_SHADERS_GLSL))
        {
          shader = cogl_create_shader (COGL_SHADER_TYPE_FRAGMENT);
          cogl_shader_source (shader, magnifier_invert_glsl_shader);
          cogl_shader_compile (shader);
          if (!cogl_shader_is_compiled (shader))
            {
              gchar *log_buf = cogl_shader_get_info_log (shader);
              g_warning (G_STRLOC ":  Unable to compile the magnifier inverse effect shader: %s",
                         log_buf);
              g_free (log_buf);

              cogl_handle_unref (shader);
              shader = COGL_INVALID_HANDLE;
            }
        }
    }

  self->shader = shader;
  self->is_attached = FALSE;
  self->tex_uniform = -1;

  if (shader != COGL_INVALID_HANDLE)
    cogl_handle_ref (self->shader);
}

/**
 * shell_mag_inverse_effect_new:
 *
 * Creates a new #ShellMagInverseEffect to be used with
 * clutter_actor_add_effect()
 *
 * Return value: the newly created #ShellMagInverseEffect or %NULL
 */
ClutterEffect *
shell_mag_inverse_effect_new (void)
{
  return g_object_new (SHELL_TYPE_MAG_INVERSE_EFFECT,
                       NULL);
}
